require "ostruct"
require "securerandom"
require "actions/errors"

module Actions
  class Context < BasicObject

    attr_reader :id
    attr_accessor :action

    def self.build(context={})
      self === context ? context : new(context)
    end

    def initialize(initial={})
      @id = ::SecureRandom.uuid
      @hash = {}
      @failure = false
      _update(initial)
    end

    def success?
      !failure?
    end

    def failure?
      @failure
    end

    def fail!(context={})
      _update(context)
      @failure = true
      ::Kernel.raise ::Actions::Errors::Failure.new(self, self)
    end

    def defined?(attribute)
      @hash.has_key?(attribute.to_s)
    end

    def called!(action)
      _called << action
    end

    def rollback!
      return false if @rolled_back
      _called.reverse_each(&:rollback)
      @rolled_back = true
    end

    def method_missing(name, *args)
      mname = name.id2name
      len = args.length
      if mname.chomp!("=")
        if len != 1
          ::Kernel.raise ::ArgumentError, "wrong number of arguments (#{len} for 1)", ::Kernel.caller(1)
        end
        @hash[mname] = args[0]
      elsif len == 0
        @hash[mname]
      else
        ::Kernel.raise ::NoMethodError, "undefined method `#{mname}' for #{self}", ::Kernel.caller(1)
      end
    end

    def respond_to?(*args)
      true
    end

    def class
      (class << self; self end).superclass
    end

    def inspect
      status = success? ? "success" : "failure"
      attributes = to_h.map { |k,v| "#{k}=#{v.inspect}" }.join(", ")
      "#{self.class.name}[#{status}](#{attributes})"
    end

    def to_s
      inspect
    end

    def to_h
      @hash
    end

    private

    def _update(context)
      context.each do |key, value|
        __send__("#{key}=", value)
      end
    end

    def _called
      @called ||= []
    end

  end
end
