require "set"
require "forwardable"
require "actions/errors"

module Actions
  class SafeContext
    extend Forwardable

    def initialize(context = {})
      @context = Context.build(context)
    end

    def input!(name, type: Object, required: false, mutable: false)
      attribute!(name, type: type, readonly: !mutable)
      validate!(name, type: type, required: required)
    end

    def output!(name, type: Object)
      attribute!(name, type: type, readonly: false)
    end

    private

    def attribute!(name, type: Object, readonly: false)
      define_singleton_method(name) do
        @context.__send__(name)
      end

      unless readonly
        define_singleton_method("#{name}=") do |value|
          unless value.is_a?(type)
            raise Errors::InvalidType.new("#{name} allows values of type #{type}, received #{value.class}", self)
          end

          @context.__send__("#{name}=", value)
        end
      end
    end

    def validate!(attribute, type: Object, required: false)
      value = @context.__send__(attribute)

      if required
        if value.nil?
          raise Errors::ValidationFailed.new("expected #{attribute} not to be nil", self)
        end
      end

      if @context.defined?(attribute)
        unless value.is_a?(type)
          raise Errors::ValidationFailed.new("expected #{attribute} to be of type #{type}, is #{value.class}", self)
        end
      end
    end

    def_delegators :@context, :success?, :failure?, :fail!, :called!,
      :rollback!, :id, :to_h, :to_s, :defined?, :inspect, :action, :action=
  end
end
